﻿Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Threading

' http://cs117-alltags-deletable.appspot.com/

''' <summary>
''' Die Klasse UrsWebDBServer stellt einen Database Web Service bereit.
''' </summary>
Public Class WebDBServer
#Region " Events "
   ''' <summary>
   ''' <b>ValueStored</b> wird ausglöst, wenn ein Wert gespeichert werden soll.
   ''' </summary>
   ''' <param name="Key">Gespeicherter Schlüssel.</param>
   ''' <param name="Value">Gespeicherte Information.</param>
   Public Event ValueStored(Key As String, Value As String)

   ''' <summary>
   ''' <b>ValueDelivered</b> wird ausgelöst, wenn eine Information abgefragt wurde.
   ''' </summary>
   ''' <param name="Key">Abgefragter Schlüssel.</param>
   ''' <param name="Value">Zurück gelieferter Wert.</param>
   Public Event ValueDelivered(Key As String, Value As String)

   ''' <summary>
   ''' <b>DBServerStarted</b> wird ausgelöst, wenn der Server gestartet wurde.
   ''' </summary>
   ''' <remarks>Bisher können nur Dateien fom Typ "text/html" abgerufen werden.</remarks>
   Public Event DBServerStarted()

   ''' <summary>
   ''' <b>DBServerStopped</b> wird ausgelöst, wenn der Server gestoppt wurde.
   ''' </summary>
   Public Event DBServerStopped()

   ''' <summary>
   ''' <b>NotFound</b> wird ausgelöst, wenn ein Request nicht bedient werden kann.
   ''' </summary>
   ''' <param name="ResourceName">Die bezeichnung der angefragten Ressource.</param>
   ''' <remarks>In diesem Fall wurde die Antwort <i>404 Not Found</i>
   ''' an den Anfragenden zurück geliefert.</remarks>
   Public Event NotFound(ResourceName As String)

   ''' <summary>
   ''' <b>RequestError</b> wird ausgelöst, wenn ein Request nicht interpretiert werden kann.
   ''' </summary>
   ''' <param name="Message">Fehlermeldung</param>
   ''' <remarks>In diesem Fall wurde die Antwort <i>400 Bad Request</i>
   ''' an den Anfragenden zurück geliefert.</remarks>
   Public Event BadRequest(Message As String)
#End Region

#Region " Felder "
   Private LocalTCPListener As TcpListener
   Private LocalPort As Integer = 80
   Private LocalAddress As IPAddress = New IPAddress(0)
   Private WebThread As Thread
   Private DataBase As ITinyWebDB
#End Region

#Region " Properties "
   ''' <summary>
   ''' Ruft den Wert für das Receive-Timeout des benutzten 
   ''' <see cref="Socket"/>-Objekts ab oder legt in fest.
   ''' </summary>
   ''' <returns>Den Wert für die Eigenschaft <see cref="Socket.ReceiveTimeout">ReceiveTimeout</see>
   ''' des benutzten <see cref="Socket"/>-Objekts</returns>
   Public Property ReceiveTimeout As Integer = 500

   ''' <summary>
   ''' Ruft einen Wert ab, der angibt, ob der Server in Betrieb ist.
   ''' </summary>
   ''' <returns></returns>
   Public ReadOnly Property IsRunning As Boolean = False
#End Region

#Region " Konstruktor "
   ''' <summary>
   ''' Initialisiert eine neue Instanz der <see cref="WebDBServer"/>-Klasse.
   ''' </summary>
   ''' <param name="DataBase">Datenbank zur Datenhaltung.</param>
   Public Sub New(DataBase As ITinyWebDB)
      Me.DataBase = DataBase
   End Sub
#End Region

#Region " Start / Stopp"
   ''' <summary>
   ''' Startet den Database Web-Service.
   ''' </summary>
   ''' <param name="LocalPort">Der lokale Port, der zum Empfang benutzt werden soll.</param>
   ''' <param name="LocalAddress">Optional. IP-Adresse des zu benutzenden Netzwerk-Interfaces (NIC).<para></para>
   ''' Wird keine Angabe gemacht, wird der Server mit "0.0.0.0" initialisiert, d.h. er nimmt 
   ''' Anfragen von alle vorhanden NICs entgegen.</param>
   Public Sub StartWebDBServer(LocalPort As Integer, Optional LocalAddress As IPAddress = Nothing)
      Me.LocalPort = LocalPort
      If Not LocalAddress Is Nothing Then
         Me.LocalAddress = LocalAddress
      End If

      LocalTCPListener = New TcpListener(Me.LocalAddress, LocalPort)
      LocalTCPListener.Start()
      WebThread = New Thread(AddressOf DoListen)
      WebThread.Start()
      RaiseEvent DBServerStarted()
   End Sub

   ''' <summary>
   ''' Stoppt den laufenden Web Service.
   ''' </summary>
   Public Sub StopWebServer()
      Try
         LocalTCPListener.Stop()
      Catch
      End Try
      Try
         WebThread.Abort()
      Catch
      End Try
      Try
         WebThread.Join()
      Catch
      End Try
   End Sub
#End Region

#Region " Respose senden"
   ''' <summary>
   ''' Sendet die Json-codierte Rückmeldung zur Datenbank-Abfrage an den Anfragenden.
   ''' </summary>
   ''' <param name="ActionCode">"STORED", "DELETED" oder "VALUE". Wird z.Zt. (2019) von der 
   ''' App Inventor App nicht ausgewertet.</param>
   ''' <param name="Tag">Schlüsselwert, Klatext, wie ursprünglich übermittelt.</param>
   ''' <param name="Value">Datenwert, Json-codiert, wie ursprünglich übermittelt.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   Protected Overridable Sub SendDBResponse(ActionCode As String, Tag As String, Value As String, Socket As Socket)
      Dim Payload = {ActionCode, Tag, Value} ' Das zurückzugebende String-Array
      Dim PayloadJson = Json.Net.JsonNet.Serialize(Payload)  ' Dessen Json-Codierung

      Dim PayloadByteCount = Encoding.UTF8.GetByteCount(PayloadJson)

      Dim Headers = "HTTP/1.1 200 OK" & vbCrLf &
          "Connection: close" & vbCrLf &
          "Content-Type: application/jsonrequest; charset=utf-8" & vbCrLf &
          "Content-Length: " & PayloadByteCount & vbCrLf &
          vbCrLf
      Dim SendData As Byte() = Encoding.UTF8.GetBytes(Headers & PayloadJson)

      Try
         Socket.Send(SendData, SendData.Length, 0)
      Catch
      End Try
   End Sub

   ''' <summary>
   ''' Sendet die Rückmeldung zur Datenbank-Abfrage als HTML-Seite an den Anfragenden.
   ''' </summary>
   ''' <param name="ActionCode">"STORED", "DELETED" oder "VALUE".</param>
   ''' <param name="Tag">Schlüsselwert, Klatext, wie ursprünglich übermittelt.</param>
   ''' <param name="Value">Datenwert, Json-codiert, wie ursprünglich übermittelt.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   ''' <remarks>Diese Methode wird aufgerufen, wenn der Request den Query-Parameter <i>fmt</i>
   ''' mit dem Wert <i>html</i> enthält.</remarks>
   Protected Overridable Sub SendHtmlResponse(ActionCode As String, Tag As String, Value As String, Socket As Socket)
      Dim Payload = {ActionCode, Tag, Value} ' Das zurückzugebende String-Array
      Dim PayloadJson = Json.Net.JsonNet.Serialize(Payload)  ' Dessen Json-Codierung

      Dim Http = "Http/1.1 200 OK" & vbCrLf &
                 "Connection: close" & vbCrLf &
                 "Content-Type: Text/html" & vbCrLf & vbCrLf &
                 "<html><body>Action: " & PayloadJson & "</body></html>"
      Dim SendData As Byte() = Encoding.UTF8.GetBytes(Http)

      Try
         Socket.Send(SendData, SendData.Length, 0)
      Catch
      End Try
   End Sub

   ''' <summary>
   ''' Sendet die HTTP-Anwort <i>400 Bad Request</i>.
   ''' </summary>
   ''' <param name="Message">Zusätzliche Angabe zu dieser Meldung.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   ''' <remarks>Es wird ein HTML-Content mit versendet. Der HTML-Body ist:<para/>
   ''' <i>"Bad Request: " + <b>Message</b></i>.<para/>
   ''' Nach dem Versand der Nachricht wird das Ereignis <see cref="BadRequest"/> ausgelöst.</remarks>
   Protected Overridable Sub SendBadRequest(Message As String, Socket As Socket)
      Dim Http = "Http/1.1 400 Bad Request" & vbCrLf &
                 "Connection: close" & vbCrLf &
                 "Content-Type: Text/html" & vbCrLf & vbCrLf &
                 "<html><body>Bad Request: " & Message & "</body></html>"
      Dim SendData As Byte() = Encoding.UTF8.GetBytes(Http)
      RaiseEvent BadRequest("Bad Request: " & Message)

      Try
         Socket.Send(SendData, SendData.Length, 0)
      Catch
      End Try
   End Sub

   ''' <summary>
   ''' Sendet die HTTP-Anwort <i>404 Not Found</i>.
   ''' </summary>
   ''' <param name="Resource">Bezeichnung der Ressource, die angefragt wurde.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   ''' <remarks>Es wird ein HTML-Content mit versendet. Der HTML-Body ist:<para/>
   ''' <i>"Not Found: " + <b>Resource</b></i>.<para/>
   ''' Nach dem Versand der Nachricht wird das Ereignis <see cref="BadRequest"/> ausgelöst.</remarks>
   Protected Overridable Sub SendNotFound(Resource As String, Socket As Socket)
      Dim Http = "Http/1.1 404 Not Found" & vbCrLf &
                 "Connection: close" & vbCrLf &
                 "Content-Type: Text/html" & vbCrLf & vbCrLf &
                 "<html><body>Not Found: " & Resource & "</body></html>"
      Dim SendData As Byte() = Encoding.UTF8.GetBytes(Http)
      RaiseEvent BadRequest("SendNotFound: " & Resource)

      Try
         Socket.Send(SendData, SendData.Length, 0)
      Catch
      End Try
   End Sub
#End Region

#Region " DB-Requests bearbeiten "
   ''' <summary>
   ''' Behandelt eine Datenbankabfrage. Die Methode wird ausgeführt, wenn der Request 
   ''' "<i>POST /getvalue</i>" ist.
   ''' </summary>
   ''' <param name="Payload">Der URL-codierte Content des HTTP-POST-Requests.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   ''' <remarks>Der reguläre Payload hat die Form <para/>
   ''' <i><![CDATA[tag=%C3%84rger&fmt=html]]></i>
   ''' <i>fmt</i> kann fehlen oder einen anderen Wert besitzen.<para/>
   ''' Ist nicht mindestens der Query-Parameter <i>tag</i> enthalten,
   ''' wird <see cref="SendBadRequest"/> aufegrufen.<para/>
   ''' Ist der Query-Paramter <i>fmt</i> mit dem Wert <i>html</i> enthalten, wird zum
   ''' Senden der Antwort die Methode <see cref="SendHtmlResponse"/> aufgerufen, ansonsten
   ''' <see cref="SendDBResponse"/>.<para/>
   ''' Die Methode wertet nur diese beiden Tags aus.<para/>
   ''' Nach dem Versand der Antwort wird das Ereignis <see cref="ValueDelivered"/> ausgelöst.
   ''' </remarks>
   Protected Overridable Sub HandleDBGetValueRequest(Payload As String, Socket As Socket)
      Dim Key As String = ""

      Dim HasTag As Boolean = False
      Dim fmt As String = ""

      Dim data = Payload.Split("&"c)

      For Each s In data
         If s.StartsWith("tag=") Then
            Key = WebUtility.UrlDecode(s.Substring(4))
            HasTag = True
            Continue For
         End If
         If s.StartsWith("fmt=") Then
            fmt = WebUtility.UrlDecode(s.Substring(4))
            Continue For
         End If
      Next

      If Not HasTag Then
         SendNotFound("getvalue no tag", Socket)
         Return
      End If

      Dim EmptyValue = """"""
      Dim Value As String = DataBase.GetValue(Key)
      If Value Is Nothing Then
         Value = EmptyValue
      End If

      If fmt = "html" Then
         SendHtmlResponse("VALUE", Key, Value, Socket)
      Else
         SendDBResponse("VALUE", Key, Value, Socket)
      End If
      RaiseEvent ValueDelivered(Key, Value)
   End Sub

   ''' <summary>
   ''' Behandelt eine Speicheranfrage. Die Methode wird ausgeführt, wenn der Request 
   ''' "<i>POST /storeavalue</i>" ist.
   ''' </summary>
   ''' <param name="Payload">Der Content des Requests.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   ''' <remarks>Der reguläre Payload hat die Form <br/>
   ''' <i><![CDATA[tag=%C3%84rger&value=%22b%C3%B6se%22&fmt=html]]></i><para/>
   ''' <b>fmt</b> kann fehlen oder einen anderen Wert besitzen.<para/>
   ''' Sind nicht mindestens die Query-Parameter <i>tag</i> und <i>value</i> enthalten,
   ''' wird <see cref="SendBadRequest"/> aufgerufen.<para/>
   ''' Hat der Parameter <i>value</i> den Wert <i>*delete*</i>, wird der entsprechende 
   ''' Datenbankeintrag gelöscht.<para/>
   ''' Ist der Query-Paramter <i>fmt</i> mit dem Wert <i>html</i> enthalten, wird zum
   ''' Senden der Antwort die Methode <see cref="SendHtmlResponse"/> aufgerufen, ansonsten
   ''' <see cref="SendDBResponse"/>.<para/>
   ''' Die Methode wertet nur diese drei Tags aus.<para/>
   ''' Nach dem Versand der Antwort wird das Ereignis <see cref="ValueStored"/> ausgelöst.
   ''' </remarks>
   Protected Overridable Sub HandleDBStoreValueRequest(Payload As String, Socket As Socket)
      Dim HasTag As Boolean = False
      Dim HasValue As Boolean = False
      Dim Key As String = ""
      Dim Value As String = ""
      Dim fmt As String = ""

      ' Query-Paramter isolieren
      Dim data = Payload.Split("&"c)
      For Each s In data
         If s.StartsWith("tag=") Then
            Key = WebUtility.UrlDecode(s.Substring(4))
            HasTag = True
            Continue For
         End If
         If s.StartsWith("value=") Then
            Value = WebUtility.UrlDecode(s.Substring(6))
            HasValue = True
            Continue For
         End If
         If s.StartsWith("fmt=") Then
            fmt = WebUtility.UrlDecode(s.Substring(4))
            Continue For
         End If
      Next

      ' Query-Paramter vollständig?
      If (Not (HasTag AndAlso HasValue)) Then
         SendBadRequest("'storeavalue' query parameter(s) missing", Socket)
         Return
      End If

      Dim Action As String
      If Value = """*delete*""" Then
         If DataBase.DeleteValue(Key) Then
            Action = "DELETED"
         Else
            Action = "DEL ERROR"
         End If
      Else
         Action = "STORED"
         DataBase.StoreValue(Key, Value)
      End If


      If fmt = "html" Then
         SendHtmlResponse(Action, Key, Value, Socket)
      Else
         SendDBResponse(Action, Key, Value, Socket)
      End If
      RaiseEvent ValueStored(Key, Value)
   End Sub

   ''' <summary>
   ''' Sendet den gesamten Datenbankinhalt als Json-String.
   ''' </summary>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   Protected Overridable Sub HandleDBGetAllRequest(Socket As Socket)
      Dim payload As String = DataBase.GetAll()
      Dim PayloadByteCount = Encoding.UTF8.GetByteCount(payload)

      Dim Headers = "HTTP/1.1 200 OK" & vbCrLf &
          "Connection: close" & vbCrLf &
          "Content-Type: application/jsonrequest; charset=utf-8" & vbCrLf &
          "Content-Length: " & PayloadByteCount & vbCrLf &
          vbCrLf
      Dim SendData As Byte() = Encoding.UTF8.GetBytes(Headers & payload)

      Try
         Socket.Send(SendData, SendData.Length, 0)
      Catch
      End Try
   End Sub
#End Region

#Region " Empfangen und analysieren "
   ''' <summary>Listen-Thread</summary>
   Private Sub DoListen()
      _IsRunning = True
      Dim ReceiveBuffer(2048) As Byte
      Try
         Do While True
            'accept new socket connection
            Dim Socket As Socket = LocalTCPListener.AcceptSocket
            If Socket.Connected Then
               Socket.ReceiveTimeout = _ReceiveTimeout
               Dim ByteCount As Integer
               Thread.Sleep(200)
               ByteCount = Socket.Receive(ReceiveBuffer, 0, ReceiveBuffer.Length, SocketFlags.None)
               Try 'evtl. ein zweites Paket
                  ByteCount += Socket.Receive(ReceiveBuffer, ByteCount, ReceiveBuffer.Length - ByteCount, SocketFlags.None)
                  ' Android-Mobil-Telefon sendet alles zusammen
               Catch
               End Try

               Dim Request As String = Encoding.UTF8.GetString(ReceiveBuffer, 0, ByteCount)
               CheckAndRespond(Request, Socket)

               Socket.Close()
            End If
         Loop
      Finally
         Try
            RaiseEvent DBServerStopped()
         Catch ex As Exception
         End Try
         _IsRunning = False
         End Try
   End Sub

   ''' <summary>
   ''' Prüft den Request und ruft die entsprechende Methode zur Beantwortung auf.
   ''' </summary>
   ''' <param name="Request">Der HTTP-Request.</param>
   ''' <param name="Socket">Socket-Objekt, über das die Rückmeldung gesendet werden soll.</param>
   ''' <remarks>Folgende Methoden zur Beantwortung der Anfrage werden aufgerufen:<br/>
   ''' <list type="table">
   '''    <listheader>  
   '''       <term>Request</term>  
   '''       <description>Methode</description>  
   '''    </listheader>  
   '''    <item>
   '''       <term><i>POST /storeavalue</i></term>  
   '''       <description><see cref="HandleDBStoreValueRequest"/></description>  
   '''    </item>  
   '''    <item>
   '''       <term><i>POST /getvalue</i></term>  
   '''       <description><see cref="HandleDBGetValueRequest"/></description>  
   '''    </item> 
   '''    <item>
   '''       <term>sonst</term>  
   '''       <description><see cref="SendNotFound"/> mit Angabe der Bezeichnung der angfragten Ressource.</description>  
   '''    </item> 
   '''</list></remarks>
   Protected Overridable Sub CheckAndRespond(Request As String, Socket As Socket)
      ' Content des Request ermitteln. Imit mit Leerzeichen von den Headern getrennt.
      Dim Payload As String = ""
      Dim LastLinePos = Request.IndexOf(vbCrLf & vbCrLf)
      If LastLinePos >= 0 Then
         Payload = Request.Substring(LastLinePos + 4)
      End If

      If Request.StartsWith("POST /storeavalue HTTP/1.1") Then
         HandleDBStoreValueRequest(Payload, Socket)
         Return
      End If

      If Request.StartsWith("POST /getvalue HTTP/1.1") Then
         HandleDBGetValueRequest(Payload, Socket)
         Return
      End If

      If Request.StartsWith("POST /getall HTTP/1.1") OrElse Request.StartsWith("GET /getall HTTP/1.1") Then
         HandleDBGetAllRequest(Socket)
         Return
      End If
      ' Nicht behandelte Ressource
      Dim url = Request.Split(vbCr(0))(0)
      SendNotFound("'" & url & "'", Socket)
   End Sub
#End Region
End Class